$ifndef _Software_UART_                     ' Create a guard against multiple includes
$define _Software_UART_
'
' Interrupt driven software serial transmit and receive UART
' The routines replace the compiler's Rsout and Rsin commands (with and without timeouts)
' The Rsin replacement utilises a serial buffer
'
' Note. The routines use a timer 0 overflow, therefore the foreground routines will be slowed down somewhat
'       Also, the higher the baud rate, the more the interrupt needs to to be called, thus the slower the foreground
'
    #Disable RSIn, Rsinto, RSOut            ' Disable the compiler's library subroutines for Rsout and Rsin (with and without timeouts)

    Include "Timers.inc"                    ' Load the timer macros into the program
    'Declare Optimiser_Level = 3             ' Uncomment for maximum optimisation (recommended)
'
' Configure the Software Serial port's details
'
    Declare Serial_Baud = 9600              ' Receiver and Transmitter baud rate (Max 38400)
    Declare Rsout_Pin = PORTC.6             ' Transmit pin
    Declare Rsout_Mode = True               ' Non-Inverted serial data (Inverted not supported)
    Declare Rsout_Pace = 100                ' Delay 0.1ms (100us) between bytes sent via Rsout
    Declare Rsin_Timeout = 1000             ' Set the Rsin timeout initially to 1 second (1000ms)
    Declare Rsin_Pin = PORTC.7              ' Receive pin
    Declare Rsin_Mode = True                ' Non-Inverted serial data (Inverted not supported)

    On_Hardware_Interrupt GoTo Serial_ISR   ' Point to the ISR handler

    Symbol cRX_Oversample 4                 ' Amount of oversampling the receiver does.
    Symbol cTimer0Value = ((Xtal * 1000000) / (4 * __Serial_Baud * cRX_Oversample)) / 2
    Symbol cMinusTimer0Value = (256 - cTimer0Value) + 4
    Symbol cTransmitNumBits = 12            ' 1 start bit + 8 data bits + 1 stop bit + safe.
'
' Receiver states
'
    Symbol cHaveNothing = 0
    Symbol cWaitHalfBit = 1
    Symbol cHaveStartBit = 2
    Symbol cWaitForStop = 3
'
' Create compiler System variables and aliases for the buffered commands
'
    Dim GPR     As Byte System                      ' Storage for WREG in tranmsitter
    Dim PP0     As Byte System                      ' \
    Dim PP0H    As Byte System                      ' / Storage for FSR0L\H registers
    Dim PP1     As Byte System                      ' \
    Dim PP1H    As Byte System                      ' / Timeout inside loop counter
    Dim _UART_wFSR0_Save As PP0.Word                ' Alias the FSR0L\H storage to PP0\H
    Dim _UART_wInsideLoop As PP1.Word               ' Alias the inside loop to PP1\H
    Dim _UART_wOutsideLoop As PRODL.Word            ' Alias the outside loop to PRODL\H registers
    Dim _UART_wFSR1_Save As Word System             ' Storage for FSR1L\H registers
    Dim _UART_bIndexIn As Byte System               ' Pointer to the next empty location in the buffer
    Dim _UART_bIndexOut As Byte System              ' Pointer to the location of the oldest byte in the buffer
    Dim _UART_wRsinTimeout As Word System           ' Holds the serial timeout value

    Dim _UART_wFSR0 As FSR0L.Word
    Dim _UART_wFSR1 As FSR1L.Word
'
' Variables used for the software serial interface
'
    Dim _UART_bFlags As Byte System
    Dim tCharacterWaiting As _UART_bFlags.0          ' 1 = A byte has been received
    Dim tSendBufferFull As _UART_bFlags.1            ' 1 = A byte is being transmitted
    Dim _UART_tTX_NextBit As _UART_bFlags.2
    Dim _UART_bReceiveBitNo As Byte System           ' Keeps track of the number of bits received
    Dim _UART_bReceiveState As Byte System           ' The current state of the receiver
    Dim _UART_bSkipSamples As Byte System            ' Used to skip receive samples
    Dim _UART_bRX_Shift As Byte System

    Dim _UART_bSendBuffer As Byte System             ' Where the byte to send is stored
    Dim _UART_bSampleCount As Byte System            ' Count (down) of the number of samples
    Dim _UART_bSendBitNo As Byte System              ' Stores which bit of the 10-bit byte is being sent
'
' Create the serial buffer
'
$ifndef Rsin_BufferSize
$define Rsin_BufferSize 16                           ' The size of the serial receive buffer (Max 255)
$endif
    Dim _UART_RingBuffer[Rsin_BufferSize] As Byte    ' Array for holding received bytes

'---------------------------------------------------------------------------------
    GoTo InitSoftwareUART                                   ' Jump over the subroutines and initialise the routines
'---------------------------------------------------------------------------------
 ' Timer0 overflow Serial Interrupt Routine
 ' Receives RS232 bytes
 ' The ISR runs cBaudRate * cRX_Oversample times per second.
 '
Serial_ISR:
    If INTCONbits_T0IF = 1 Then                             ' Is this a Timer0 interrupt?
        TMR0L = cMinusTimer0Value                           ' Yes. So reset Timer0 value
        INTCONbits_T0IF = 0                                 ' Clear the Timer0 overflow bit
        Dec _UART_bSkipSamples
        If _UART_bSkipSamples = 0 Then
            Inc _UART_bSkipSamples                          ' Check next time
            Select _UART_bReceiveState                      ' Operate on the appropriate receive state
                Case cHaveNothing
                    If __Rsin_Port.__Rsin_Pin = 0 Then      ' Check for start bit of a received byte
                        _UART_bSkipSamples = cRX_Oversample / 2
                        Inc _UART_bReceiveState             ' Move to the next receiver state
                    EndIf
                Case cWaitHalfBit
                    If __Rsin_Port.__Rsin_Pin = 0 Then      ' Valid start bit ?
                        _UART_bReceiveBitNo = 8             ' Yes. So setup for 8-data bits
                        _UART_bSkipSamples = cRX_Oversample ' Setup the oversample value
                        Inc _UART_bReceiveState             ' Move to the next receiver state
                    Else                                    ' Otherwise...
                        _UART_bReceiveState = cHaveNothing  ' Reset the receive state
                    EndIf
                Case cHaveStartBit                          ' Did we get a start bit?
                    _UART_bRX_Shift = _UART_bRX_Shift >> 1  ' Yes. So start collecting data bits
                    _UART_bRX_Shift.7 = __Rsin_Port.__Rsin_Pin
                    _UART_bSkipSamples = cRX_Oversample
                    Dec _UART_bReceiveBitNo
                    If _UART_bReceiveBitNo = 0 Then         ' Last bit ?
                        Inc _UART_bReceiveState             ' Yes. So move to the next receiver state
                    EndIf
                Case cWaitForStop                           ' Stop bit detected ?
                    _UART_wFSR1_Save = _UART_wFSR1          ' Yes. So save FSR1L\H registers
                    Inc _UART_bIndexIn                      ' Move up the buffer
                    If _UART_bIndexIn >= Rsin_BufferSize Then ' End of buffer reached ?
                        _UART_bIndexIn = 0                  ' Yes. So reset it
                    EndIf
                    _UART_wFSR1 = VarPtr _UART_RingBuffer   ' Point FSR1L\H to _UART_RingBuffer
                    _UART_wFSR1 = _UART_wFSR1 + _UART_bIndexIn ' Add the buffer position to FSR1L\H
                    INDF1 = _UART_bRX_Shift                 ' Place the received byte into the buffer
                    _UART_wFSR1 = _UART_wFSR1_Save          ' Restore FSR1L\H registers
                    tCharacterWaiting = 1                   ' Signal a byte has been received
                    _UART_bReceiveState = cHaveNothing      ' Reset the state flag to wait for another byte
                    _UART_wFSR1 = _UART_wFSR1_Save          ' Restore the contents of FSR1L/H
            End Select
        EndIf
        '
        ' Transmit
        ' This will be called every cRX_Oversample time (because the receive needs to over-sample the incoming data)
        '
        If tSendBufferFull = 1 Then
            Dec _UART_bSampleCount
            If _UART_bSampleCount = 0 Then
                _UART_bSampleCount = cRX_Oversample
                __Rsout_Port.__Rsout_Pin = _UART_tTX_NextBit ' Send next bit
                _UART_tTX_NextBit = _UART_bSendBuffer.0
                _UART_bSendBuffer = _UART_bSendBuffer >> 1
                _UART_bSendBuffer.7 = 1
                Dec _UART_bSendBitNo
                If _UART_bSendBitNo == 0 Then
                    tSendBufferFull = 0                     ' Clear the flag once finished
                EndIf
            EndIf
        EndIf
    EndIf
$ifdef _Hardware_USART_                                     ' Are the Hardware routines also being used?
    GoTo _Usart_High_Prior_Interrupt                        ' Yes. So jump to the hardware USART interrupt routine
$else                                                       ' Otherwise...
    retfie fast                                             ' Exit the interrupt, restoring the WREG, BSR and STATUS registers
$endif
'------------------------------------------------------------------------------
' Wait for a byte from the interrupt driven circular buffer with timeout
'
' Input     : Word variable _UART_wRsinTimeout holds the timeout value (in ms)
' Output    : PP0 and WREG hold the value received
'           : Carry Flag (STATUS.0) is clear if timed out
' Notes     : Uses PRODL\H as temporary variable storage
'           : Uses FSR0L\H as buffer pointers
'           : Replaces the compiler's Rsin command with timeouts
'
#ifdef Rsinto#Req
RSIn2:
@Rsin#Buffer_Req = 1                                ; Enable the _Uart_GetByte subroutine
    _UART_wOutsideLoop = _UART_wRsinTimeout         ' Transfer the timeout variable to it's working variable
    Clear _UART_wInsideLoop                         ' Reset the inside loop counter
_UART_Getbuf_Outside:
    bra $ + 2                                       ' Delay for 2 cycles within the outside loop
_UART_Getbuf_Inside:
    #ifdef Watchdog_Req                             ' Delay for 1 cycle within the inside loop
        clrwdt
    #else
        nop
    #endif
    movf   _UART_bIndexIn,w                         ' \
    subwf  _UART_bIndexOut,w                        ' / Is there anything in the serial buffer?
    bnz    _Uart_GetByte                            ' Yes. So fetch it
    WREG = 255                                      ' \
    addwf  _UART_wInsideLoop,f                      '  |
    addwfc _UART_wInsideLoopH,f                     '   Decrement the inside and outside loops
    addwfc _UART_wOutsideLoop,f                     '  |
    addwfc _UART_wOutsideLoopH,f                    ' /
    btfss     STATUS,C
    ret                                             ' Return with the Carry flag clear to indicate timed out
    infsnz _UART_wInsideLoop,w
    incfsz _UART_wInsideLoopH,w
    bra    _UART_Getbuf_Outside
    _UART_wInsideLoop = ((59 * Xtal) / 4)           ' Set the inside loop counter based upon the Xtal frequency
    bra    _UART_Getbuf_Inside
#endif
'------------------------------------------------------------------------------
' Wait for a byte from the interrupt driven circular buffer without timeout
'
' Input     : None
' Output    : bGetCh holds the value received
' Notes     : Uses FSR0L\H as buffer pointers
'           : Replaces the compiler's Rsin command (without timeout)
'
#ifdef Rsin#Req
@Rsin
@Rsin#Buffer_Req = 1                                ; Enable the _Uart_GetByte subroutine
    While _UART_bIndexIn = _UART_bIndexOut          ' Wait for a byte to appear in the serial buffer
    #ifdef Watchdog_Req
        clrwdt                                      ' Clear the watchdog timer within the loop (if required)
    #endif
    Wend
#endif
' Fall through to _Uart_GetByte
'------------------------------------------------------------------------------
' Read a byte from the interrupt driven circular buffer
'
' Input     : None
' Output    : WREG and PP0 hold the value received
' Notes     : Uses FSR0L\H as buffer pointers
'
#ifdef Rsin#Buffer_Req
_Uart_GetByte:
    Inc _UART_bIndexOut                             ' Increment _UART_bIndexOut pointer (0 to 255)
    If _UART_bIndexOut >= Rsin_BufferSize Then      ' End of buffer reached?
        _UART_bIndexOut = 0                         ' Yes. So clear _UART_bIndexOut
    EndIf
    _UART_wFSR0_Save = _UART_wFSR0                  ' Save FSR0L\H registers
    _UART_wFSR0 = VarPtr _UART_RingBuffer           ' Point FSR0L\H to _UART_RingBuffer
    _UART_wFSR0 = _UART_wFSR0 + _UART_bIndexOut     ' Add the buffer position to FSR0L\H
    WREG = INDF0                                    ' Read buffer location (_UART_bIndexOut) into WREG
    PP0 = WREG                                      ' Also place it into PP0
    _UART_wFSR0 = _UART_wFSR0_Save                  ' Restore FSR0\H registers
    STATUSbits_C = 1                                ' Set the Carry flag to indicate a byte received
    Return
#endif
'---------------------------------------------------------------------------------
' Transmit a byte
'
' Input     : WREG is the byte to transmit
' Output    : tSendBufferFull is clear when the byte has been transmitted
' Notes     : WREG is unchanged when the routine returns
'           : Replaces the compiler's Rsout command
'
#ifdef Rsout#Req
@Rsout
    GPR = WREG                                      ' Save the contents of WREG for later
    While tSendBufferFull = 1                       ' Wait for the previous byte to be transmitted
    #ifdef Watchdog_Req
        clrwdt                                      ' Clear the watchdog timer within the loop (if required)
    #endif
    Wend
    _UART_tTX_NextBit = 0                           ' Reset the bit
    _UART_bSampleCount = 1
    _UART_bSendBuffer = GPR                         ' Place the byte to send in the buffer
    _UART_bSendBitNo = cTransmitNumBits             ' The amount of bits to send
    tSendBufferFull = 1                             ' Signal a byte is being transmitted
#if(__Rsout_Pace > 0)                               ' Is a delay required after each byte sent?
    #ifdef dlus@w                                   ' Yes. So is the DelayUs routine inmemory?
    PP0H = __Rsout_Pace.Byte1                       ' Yes. So initiate a microsecond delay
    WREG = __Rsout_Pace.Byte0
    Call(dlus@w)                                    ' Call the DelayUs library subroutine
    #endif
#endif
    WREG = GPR                                      ' Restore the contents of WREG
    Return
#endif
'---------------------------------------------------------------------------------
' Initialise the software UART
'
' Input     : None
' Output    : None
' Notes     : Sets up the the receive pin as an input
'             Sets up the the transmit pin as an output
'             Sets up an 8-bit Timer0 overflow interrupt
'
$define InitUart() GoSub _Init_uart
_Init_uart:
    _UART_wRsinTimeout = __Rsin_Timeout             ' Set the timeout value from the "Serial_Baud" Declare
    Clear _UART_RingBuffer
    _UART_bIndexIn = 0
    _UART_bIndexOut = 0
    _UART_bReceiveState = cHaveNothing              ' Reset the receive state
    _UART_bSkipSamples = 1                          ' Check each interrupt for start bit
    tCharacterWaiting = 0
    Input __Rsin_Port.__Rsin_Pin                    ' RX pin to onput
    Output __Rsout_Port.__Rsout_Pin                 ' TX pin to output
'
' Configure Timer 0 for:
'                       Interrupt on Timer0 overflow
'                       8-bit operation
'                       Internal clock source
'                       1:2 Prescaler
'
    OpenTimer0(TIMER_INT_ON & T0_8BIT & T0_SOURCE_INT & T0_PS_1_2)
    TMR0L = cTimer0Value                            ' Initialise the value of Timer0
    INTCONbits_GIE = 1                              ' Enable global interrupts
    Return
'--------------------------------------------------------------------------------
' Clear the Serial Buffer
'
' Input     : None
' Output    : None
' Notes     : Also resets the index pointers to the serial buffer
'
$define ClearSerialBuffer() _ClearSerialBuffer
_ClearSerialBuffer Macro
    GoSub __Clear_Uart_Buffer
    Endm
#ifdef _ClearSerialBuffer#Req
__Clear_Uart_Buffer:
    INTCONbits_T0IE = 0                             ' Disable interrupt on Timer0 overflow
    Clear _UART_RingBuffer                          ' Clear the serial buffer
    _UART_bIndexIn = 0                              ' Clear the buffer internal pointer
    _UART_bIndexOut = 0                             ' Clear the buffer external pointer
    INTCONbits_T0IE = 1                             ' Re-Enable interrupt on Timer0 overflow
    Return
#endif
'---------------------------------------------------------------------------------
' Set the Timeout of Rsin, but only if the timeout version of Rsin is used
$define SetTimeout(pTime) '
#ifdef Rsinto#Req         '
    _UART_wRsinTimeout = pTime  '
#endif
'---------------------------------------------------------------------------------
' Initialise the Software UART
InitSoftwareUART:
    InitUart()

$endif

